@@ -14,6 +14,9 @@ from homeconnect_webthing.auth import Auth
 
 class EventListener(ABC):
 
+    def on_connected(self):
+        pass
+
     def on_keep_alive_event(self, event):
         pass
 
@@ -23,6 +26,9 @@ class EventListener(ABC):
     def on_status_event(self, event):
         pass
 
+    def on_event_event(self, event):
+        pass
+
 
 class Device(EventListener):
 
@@ -56,14 +62,18 @@ class Device(EventListener):
     def _perform_put(self, path:str, data: str, num_tries: int = 1):
         uri = self._uri + path
         for i in range(0, num_tries):
-            logging.info("query PUT " + uri)
+            if i > 0:
+                logging.info("query PUT " + uri + " (" + str(i) + ". retry)")
+            else:
+                logging.info("query PUT " + uri)
+
             response = requests.put(uri, data=data, headers={"Content-Type": "application/json", "Authorization": "Bearer " + self._auth.access_token})
             if self.__is_success(response.status_code):
                 return
             else:
                 logging.warning("error occurred by calling POST " + uri + " " + data)
-                logging.warning("got " + response.text)
-                sleep(2)
+                logging.warning("got " + str(response.status_code) + " " + str(response.text))
+                sleep(1 + num_tries)
         raise Exception("error occurred by calling GET " + uri + " Got " + str(response.status_code) + " " + str(response.text))
 
     @property
@@ -90,7 +100,6 @@ class Device(EventListener):
 class Dishwasher(Device):
 
     def __init__(self, uri: str, auth: Auth, name: str, device_type: str, haid: str, brand: str, vib: str, enumber: str):
-        self.date_refreshed = datetime.now()
         self.__power = ""
         self.__operation = ""
         self.__door = ""
@@ -113,20 +122,38 @@ class Dishwasher(Device):
 
     def register_value_changed_listener(self, value_changed_listener):
         self._value_changed_listeners.add(value_changed_listener)
+        logging.info("value_changed_listener " + str(value_changed_listener) + " registered")
+        self.__notify_listeners()
+
+    def __notify_listeners(self):
+        for value_changed_listener in self._value_changed_listeners:
+            value_changed_listener()
+
+    def on_connected(self):
+        logging.info("refresh state (new event stream connection)")
+        self.__refresh()
+
+    def on_keep_alive_event(self, event):
+        for value_changed_listener in self._value_changed_listeners:
+            value_changed_listener()
 
     def on_notify_event(self, event):
+        logging.debug("notify event: " + str(event.data))
         self.on__value_changed_event(event)
 
     def on_status_event(self, event):
+        logging.debug("status event: " + str(event.data))
         self.on__value_changed_event(event)
 
+    def on_event_event(self, event):
+        logging.debug("event event: " + str(event.data))
+
     def on__value_changed_event(self, event):
         if event.id == self.haid:
             try:
                 data = json.loads(event.data)
                 self.__on_value_changes(data.get('items', []))
-                for value_changed_listener in self._value_changed_listeners:
-                    value_changed_listener()
+                self.__notify_listeners()
             except Exception as e:
                 logging.warning("error occurred by handling event " + str(event), e)
 
@@ -134,12 +161,16 @@ class Dishwasher(Device):
         for record in changes:
             if record['key'] == 'BSH.Common.Status.DoorState':
                 self.__door = record['value']
+                logging.info("door state updated " + str(self.__door))
             elif record['key'] == 'BSH.Common.Status.OperationState':
                 self.__operation = record['value']
+                logging.info("operation state updated " + str(self.__operation))
             elif record['key'] == 'BSH.Common.Status.RemoteControlStartAllowed':
                 self.remote_start_allowed = record['value']
+                logging.info("remote start allowed updated " + str(self.remote_start_allowed))
             elif record['key'] == 'BSH.Common.Setting.PowerState':
                 self.__power = record['value']
+                logging.info("power state updated " + str(self.__power))
             elif record['key'] == 'BSH.Common.Root.SelectedProgram':
                 self.__program_selected = record['value']
             elif record['key'] ==  'BSH.Common.Root.ActiveProgram':
@@ -161,11 +192,6 @@ class Dishwasher(Device):
             else:
                 logging.info("unknown changed " + str(record))
 
-    def on_keep_alive_event(self, event):
-        if datetime.now() > (self.date_refreshed + timedelta(minutes=45)):
-            logging.info("refresh state")
-            self.__refresh()
-
     def __refresh(self):
         try:
             self.__on_value_changes(self._perform_get('/settings')['data']['settings'])
@@ -173,9 +199,7 @@ class Dishwasher(Device):
             record = self._perform_get('/programs/selected')['data']
             self.__program_selected = record['key']
             self.__on_value_changes(record['options'])
-            self.date_refreshed = datetime.now()
-            for value_changed_listener in self._value_changed_listeners:
-                value_changed_listener()
+            self.__notify_listeners()
         except Exception as e:
             logging.warning("error occurred on refreshing", e)
 
@@ -210,21 +234,14 @@ class Dishwasher(Device):
         else:
             return ""
 
-    def to_period(self, future_date: str):
-        remaining_secs_to_wait = (datetime.fromisoformat(future_date) - datetime.now()).seconds
-        if remaining_secs_to_wait > (60 * 60):
-            return int(remaining_secs_to_wait / 60), "minutes"
-        else:
-            return remaining_secs_to_wait, "seconds"
-
-
     def set_start_date(self, dt: str):
         if self.operation == "BSH.Common.EnumType.OperationState.Run":
             logging.info("dishwasher is already running")
         else:
-            dz = datetime.now()
-            delay = (datetime.fromisoformat(dt) - datetime.now())
             remaining_secs_to_wait = int((datetime.fromisoformat(dt) - datetime.now()).total_seconds())
+            if remaining_secs_to_wait < 0:
+                logging.warning("negative delay " + str(remaining_secs_to_wait) + " (start date: " + dt + ") delay set to 5 sec")
+                remaining_secs_to_wait = 5
             if remaining_secs_to_wait > 86000:
                 logging.warning("large delay " + str(remaining_secs_to_wait) + " (start date: " + dt + ") reduced to 86000")
                 remaining_secs_to_wait = 86000
@@ -238,9 +255,12 @@ class Dishwasher(Device):
                                 } ]
                 }
             }
-            self._perform_put("/programs/active", json.dumps(data, indent=2), num_tries=3)
-            logging.info("dishwasher program " + self.program_selected + " starts in " + str(remaining_secs_to_wait) + " secs")
-            self.__refresh()
+            try:
+                self._perform_put("/programs/active", json.dumps(data, indent=2), num_tries=3)
+                logging.info("dishwasher program " + self.program_selected + " starts in " + str(remaining_secs_to_wait) + " secs")
+                self.__refresh()
+            except Exception as e:
+                logging.warning("error occured by starting dishwasher", e)
 
     def __str__(self):
         return "power=" + str(self.power) + \
@@ -272,17 +292,31 @@ class HomeConnect:
     def __listening_for_events(self):
         sleep(3)
         uri = HomeConnect.API_URI + "/homeappliances/events"
-        num_reconnect = 0
+        num_reconnects = 0
         while True:
             try:
-                num_reconnect += 1
-                logging.info("opening sse socket to " + uri)
-                response = requests.get(uri, stream=True, headers={'Accept': 'text/event-stream', "Authorization": "Bearer " + self.auth.access_token})
-                response.raise_for_status()
+                self.__consume_sse_events(uri, max_connection_time_minutes=55)
+                num_reconnects = 0
+            except Exception as e:
+                logging.warning("Event stream (" + uri + ") error: " + str(e))
+                wait_time_sec = {0: 3, 1:5, 2: 30, 3: 2*60, 4: 5*60}.get(num_reconnects, 20*60)
+                num_reconnects += 1
+                logging.info("try " + str(num_reconnects) + ". reconnect in " + str(wait_time_sec) + " sec")
+                sleep(wait_time_sec)
+
+    def __consume_sse_events(self, uri: str, max_connection_time_minutes: int):
+        client = None
+        try:
+            logging.info("opening event stream connection " + uri)
+            response = requests.get(uri, stream=True, headers={'Accept': 'text/event-stream', "Authorization": "Bearer " + self.auth.access_token})
+            if response.status_code == 200:
                 client = sseclient.SSEClient(response)
+                logging.info("consuming events...")
+
+                for notify_listener in self.notify_listeners:
+                    notify_listener.on_connected()
+                connect_time = datetime.now()
 
-                num_reconnect = 0
-                logging.info("consuming event...")
                 for event in client.events():
                     if event.event == "NOTIFY":
                         for notify_listener in self.notify_listeners:
@@ -293,13 +327,21 @@ class HomeConnect:
                     elif event.event == "STATUS":
                         for notify_listener in self.notify_listeners:
                             notify_listener.on_status_event(event)
+                    elif event.event == "Event":
+                        for notify_listener in self.notify_listeners:
+                            notify_listener.on_event_event(event)
                     else:
                         logging.info("unknown event type " + str(event.event))
-            except Exception as e:
-                logging.warning("Error occurred by opening sse socket to " + uri + " " + str(e))
-                wait_time_sec = {0: 3, 1:5, 2: 30, 3: 2*60, 4: 5*60}.get(num_reconnect, 30*60)
-                logging.info("try reconnect in " + str(wait_time_sec) + "sec")
-                sleep(wait_time_sec)
+                    # max connection time reached?
+                    if datetime.now() > (connect_time + timedelta(minutes=max_connection_time_minutes)):
+                        logging.info("closing event stream. Max connect time " + str(max_connection_time_minutes) + " min reached (periodic reconnect)")
+                        return
+            else:
+                logging.warning("error occurred by opening event stream connection " + uri)
+                logging.warning("got " + str(response.status_code) + " " + response.text)
+        finally:
+            if client is not None:
+                client.close()
 
     def devices(self) -> List[Device]:
         uri = HomeConnect.API_URI + "/homeappliances"
