@@ -20,11 +20,7 @@ class DishwasherThing(Thing):
             description
         )
 
-        self.ioloop = tornado.ioloop.IOLoop.current()
-
         self.dishwasher = dishwasher
-        dishwasher.register_value_changed_listener(self.on_value_changed)
-
 
         self.name = Value(dishwasher.name)
         self.add_property(
@@ -221,6 +217,10 @@ class DishwasherThing(Thing):
                          'readOnly': True,
                      }))
 
+        self.ioloop = tornado.ioloop.IOLoop.current()
+        self.dishwasher.register_value_changed_listener(self.on_value_changed)
+
+
     def on_value_changed(self):
         self.ioloop.add_callback(self.__on_value_changed)
 
